import static java.lang.Math.*;
public class atom{
    /*Properties
    ______________________________________*/
    vector pos,             // current position
           force,           // force accumulator
           vel;             // velocity
     double radius,         // contact radius
           mass,            // atom mass >= 1
           imass;           // inverse of mass precalc
    boolean rested;         // if moving or not

    static vectorFunc v = new vectorFunc();

    //spring
    static int maxLink = 6;
    int linkedTo[] = new int[maxLink];
    int linkQty = 0;
    double lenght[] = new double[maxLink],
           rigidity[] = new double[maxLink],
           breaklimit[] = new double[maxLink];
    
    /*Constructor
    _____________________________________*/
    public atom(atom a)
    {
        pos = a.pos;
        vel = a.vel;
        force = a.force;
        radius = a.radius;
        mass = a.mass;
        imass = a.imass;
        rested = a.rested;
        //spr
        linkQty = a.linkQty;
        linkedTo = a.linkedTo;
        lenght = a.lenght;
        rigidity = a.rigidity;
        breaklimit = a.breaklimit;
    }
    
    public atom(int x, int y)
    {
        pos = new vector((double)x, (double)y);
        radius = 2.0d;
        mass = 10.0d;
        imass = 1.0d/mass;
        vel = new vector();
        force = new vector();
        rested = false;
        //spr
        linkQty = 0;
    }
    
    public atom(int x, int y, double numass, double nuradius)
    {
        pos = new vector((double)x, (double)y);
        radius = nuradius;
        mass = numass;
        imass = 1.0d/mass;
        vel = new vector(0,0);
        force = new vector();
        rested = false;
        //spr
        linkQty = 0;
    }
    
    /*Method
    ____________________________________________________*/
    
    public void zeroforce()    // reset force accumulator
    {
        force = new vector(0,0);    
    }
    
    public void restingTest()  // test if moving, return bool
    {
        if (vel.x >= -10 && vel.x <= 10 && vel.y >= -10 && vel.y <= 10)
            rested = true;
        else
            rested = false;
    }
    
    public boolean getrest()   // return rested state
    {        
        return rested;
    }
    
    public void unrest()       // set atom unrested
    {
        rested = false;
    }
    
    public void rested()       // set atom rested
    {
        rested = true;
    }
    
    public double distanceCenters(atom j)    // find distance between atoms (center to center)
    {
        return v.mag(v.sub(pos, j.pos));
    }
    
    public double distanceCentersSQ(atom j)    // find distanceSQ between atoms (center to center)
    {
        return v.magSQ(v.sub(pos, j.pos));
    }
    
    public double distanceSurfaces(atom j)    // find distance between atoms (from surface to surface)
    {
        return v.mag(v.sub(pos, j.pos)) - sizes(j);
    }
    
    public double distanceSurfacesSQ(atom j)    // find distanceSQ between atoms (from surface to surface)
    {
        return v.magSQ(v.sub(pos, j.pos)) - sizes(j);
    }
    
    public vector separate()        // return randomized separation force adjusted for mass
    {
        double rx = (double)Math.random()+0.5d;
        double ry = (double)Math.random()-0.5d;
        return v.scale(new vector(rx, ry), rx+ry*1000*mass);
    }
    
    public void addforce(vector f)    // accumulate force
    {
        force = v.add(force, f);
    }
    
    public void subforce(vector f)    // negative accumulate force
    {
        force = v.sub(force, f);
    }
    
    public double sizes(atom j)    // return combined radius of two atoms
    {
        return radius+j.radius;
    }
                                           
    public vector repulse(atom j, double restlen, double distcent)    // return self repulsion spring+damper forces
    {
        vector link = v.sub(pos, j.pos);
        double mult = restlen /  ((double)sqrt(distcent) );
        //double mult =  (distcent - restlen)*3000000;
        double comp = (mass*j.mass)/(mass+j.mass);
        
        vector spring = v.scale(v.sub(v.scale(link, mult), link), comp*250000); // dist=distance surface to surface. (how deep) K=5000
        vector damper = v.scale((v.sub(j.vel, vel)), comp*50);            // D=damping
        return v.add(spring, damper);
    }
    
    public void gravity(vector grav)        // apply gravity on atom
    {
        force = v.add(force, v.scale(grav, mass));
    }
    
    public void integrate(double dt)        // integrate system
    {
        vel = v.add(vel, v.scale(v.scale(force, imass), dt));     // v = (f/m) *dt   or   v = a*dt  or  v = integrated v
        pos = v.add(pos, v.scale(vel, dt));                           // x += vel *dt
    }
    
    public void hardcollision(vector boxSize, vector boxPos, double decay, double friction) // only for 0' and 90' angles
    {
        if (pos.x-radius < boxPos.x)     //else if can miss corner collisions
        {
            pos = new vector(boxPos.x+radius, pos.y);                 // adjust position
            vel = new vector(vel.x * -decay, vel.y * friction);       // adjust velocity with friction and wall bounce factors
        }
        if (pos.x+radius > boxPos.x+boxSize.x)
        {
            pos = new vector(boxPos.x+boxSize.x-radius, pos.y);
            vel = new vector(vel.x * -decay, vel.y * friction);
        }
        if (pos.y-radius < boxPos.y)
        {
            pos = new vector(pos.x, boxPos.y+radius);
            vel = new vector(vel.x * friction, vel.y * -decay);
        }
        if (pos.y+radius > boxPos.y+boxSize.y)
        {
            pos = new vector(pos.x, boxPos.y+boxSize.y-radius);
            vel = new vector(vel.x * friction, vel.y * -decay);
        }
    }
    
    public void terminalforces()        // limit maximum forces in atom accumulator
    {
        if (force.x > 8000000)
        {
            force = new vector(8000000,force.y);
        }
        if (force.x < -8000000)
        {
            force = new vector(-8000000,force.y);
        }
        if (force.y > 8000000)
        {
            force = new vector(force.x,8000000);
        }
        if (force.y < -8000000)
        {
            force = new vector(force.x,-8000000);
        }
    }
    
    public void terminalvelocity()        // limit maximum velocity
    {
        if (vel.x > 10000)
        {
            vel = new vector(10000,vel.y);
        }
        if (vel.x < -10000)
        {
            vel = new vector(-10000,vel.y);
        }
        if (vel.y > 10000)
        {
            vel = new vector(vel.x,10000);
        }
        if (vel.y < -10000)
        {
            vel = new vector(vel.x,-10000);
        }
    }
    
    public void velocitydrag()        // apply scaling factor to velocity, simulate drag
    {
        vel = v.scale(vel, 0.998d);
    }
    
    public double getmass()    // get mass
    {       
        return mass;
    }
    
    public vector getpos()    // get pos
    {       
        return pos;
    }
    
    public vector getvel()    // get vel
    {       
        return vel;
    }
    
    public double getsize()    // get size
    {       
        return radius;
    }
    
    //spring
    
    public void linkWith(int atomNum, double length)
    {
        if (linkQty >= maxLink-1)
            return;
        linkedTo[linkQty] = atomNum;
        breaklimit[linkQty] = 30000;
        rigidity[linkQty] = 30;
        lenght[linkQty] = length;
        linkQty++;
    }
    
    public void springForce(atom[] atomLst)
    {
        if (linkQty == 0)
            return;
        atom atomDest;
        for (int i=0; i<linkQty;i++)
        {
            atomDest = atomLst[linkedTo[i]];
            double distCenters = (double)sqrt(distanceCentersSQ(atomDest));  // returns SQdistance of atoms centers
            if (distCenters == 0)
            {
                distCenters = 0.1d;                     // case distance = 0, make it work. (could delete spring too or break)
            }
            vector link = v.sub(pos, atomDest.pos);                // get vector between i and j
            double mult =  (distCenters - lenght[i])*-25000;                         // get force multiple: -K*( dist(i,j) - spring.lenght)
            double comp = (mass*atomDest.mass)/(mass+atomDest.mass);      // get mass compensation
            
            vector spring = v.scale(v.sub(v.scale(link, mult), link), comp*0.7);      // calculate spring forces
            vector damper = v.scale((v.sub(atomDest.vel, vel)), comp*50);                   // D=damping  (semi broken) 
         
            addforce(v.add(spring, damper));
        }
    }
};
